/***********************************************************************
 *            main.c
 *
 *  Tue Jun 15 01:34:05 2004
 *  Copyright  2004  Michael Rasmussen
 *  mir@miras.org
 **********************************************************************/


/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */
 
#include <stdio.h>
#include <unistd.h>
#include <popt.h>
#include "socklib.h"
#include "utils.h"
#include "workers.h"

#define _REENTRANT

void usage(int exitcode, char *error, char *addl);
void setup(pthread_attr_t *attrp);

void usage(int exitcode, char *error, char *addl) {
	fprintf(stderr, "Usage: webserver [options] <port>\n"
					"[options] include:\n"
					"\t-d\tRun webserver in deamon mode\n"
					"\t-h\tShow this help\n"
					"\t-r\troot folder for webserver\n"
					"\t-D\tRun webserver in debug mode\n");
	if (error) fprintf(stderr, "%s: %s\n", error, addl);
	exit(exitcode);
}

void setup(pthread_attr_t *attrp) {
	pthread_attr_init(attrp);
	pthread_attr_setdetachstate(attrp, PTHREAD_CREATE_DETACHED);
	
	time(&server_started);
	server_requests = 0;
	server_bytes_sent = 0;
}

int main(int argc, char *argv[]) {
	int				sock, client;
	int				*fdptr;
	int				port;
	pthread_t		worker;
	pthread_attr_t	attr;
	char 			*root;
	poptContext		optCon;
	char 			c;
	struct poptOption optionsTable[] = {
		{ "daemon", 'd', POPT_ARG_NONE, 0, 'd' },
		{ "help", 'h', POPT_ARG_NONE, 0, 'h' },
		{ "root", 'r', POPT_ARG_STRING, &root, 0 },
		{ "debug", 'D', POPT_ARG_NONE, 0, 'D' },
		{ NULL, 0, 0, NULL, 0 }
	};
		
	deamon = debug = 0;
	mes = (char *) malloc(sizeof(char) * stringbuf + 1);
	if (argc < 2) usage(1, "Not enough arguments", "");
	port = atoi(argv[argc - 1]);
	if (port < 1 || port > 65535)
		usage(2, "port number must be: 0 < port < 65536", argv[argc - 1]);

	optCon = poptGetContext("webserver", argc, (const) argv, optionsTable, 0);
	
	while ((c = poptGetNextOpt(optCon)) >= 0) {
		switch (c) {
			case 'd':
				deamon = 1;
				break;
			case 'h':
				usage(0, NULL, NULL);
				break;
			case 'D':
				debug = 1;
				break;
		}
	}
	
	if (c < -1) {
		sprintf(mes, "%s: %s\n",
			poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
			poptStrerror(c));
		errorlog(mes);
	}
	poptFreeContext(optCon);
				
	getowner(argv[0]);
	
	if (root) {
		if (chdir(root) < 0)
			errorlog("chdir");
	}
	else {
		root = (char *) malloc(sizeof(char) * stringbuf + 1);
		if (getcwd(root, stringbuf) == NULL)
			errorlog("getcwd");
	}

	if (deamon)
		deamonize();
	
	sprintf(mes, "Waiting for calls on port #%d\n"
					"Using this folder for root: %s\n",
					port, root);
	infolog(mes);
	sock = make_server_socket(port);
	if (sock == -1 && pid == 0) {
		errorlog("Terminating: Error making socket");
	}
	
	setup(&attr);

	while(1) {
		client = accept(sock, NULL, NULL);
		server_requests++;
		
		fdptr = malloc(sizeof(int));
		*fdptr = client;
		pthread_create(&worker, &attr, handle_call, fdptr);
		free(fdptr);
	}
}
