/***************************************************************************
 *            utils.c
 *
 *  Tue Jun 29 19:44:27 2004
 *  Copyright  2004  Michael Rasmussen
 *  mir@miras.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "utils.h"
#include "workers.h"

char	*rfc1945_commands[] = {
		"HEAD",
		"GET"
};


void sanitize(char *str) {
	char	*src, *dest;
	
	
	src = dest = str;
	while(*src) {
		if (strncmp(src, "/../", 4) == 0)
			src +=3;
		else if (strncmp(src, "//", 2) == 0)
			src++;
		else
			*dest++ = *src++;
	}
	*dest = '\0';
	if (*str == '/')
		strcpy(str, str+1);
	if (str[0] == '\0' || strcmp(str, "./") == 0 || strcmp(str, "./..") == 0)
		strcpy(str, ".");
}

char *standardHeader(int code, char *msg, char *type) {
	struct tm 	*now;
	time_t		clock;
	char		timestr[80];
	char		*header, tmp[stringbuf];
	
	header = (char *) malloc(sizeof(char) * stringbuf + 1);
	clock = time(NULL);
	now = gmtime(&clock);
	strftime(timestr, sizeof(timestr)-1, "%a, %d %b %Y %H:%M:%S %Z", now);
	sprintf(tmp, "HTTP/1.0 %d %s\r\n", code, msg);
	strcpy(header, tmp);
	sprintf(tmp, "Date: %s\r\n", timestr);
	strcat(header, tmp);
	sprintf(tmp, "Server: Simplified threaded http server\r\n");
	strcat(header, tmp);
	sprintf(tmp, "Connection: close\r\n");
	strcat(header, tmp);
	if (type) {
		sprintf(tmp, "Content-type: %s\r\n\r\n", type);
		strcat(header, tmp);
	}
	return header;
}

int http_reply(int fd, FILE **fpp, int code, char *msg, char *type, char *content) {
	FILE	*fp = fdopen(fd, "w");
	int		bytes = 0;
	
	if (fp != NULL) {
		bytes = fprintf(fp, standardHeader(code, msg, type));
		if (content)
			bytes += fprintf(fp, "%s\r\n", content);
	}
	fflush(fp);
	if (fpp)
		*fpp = fp;
	else
		fclose(fp);
	return bytes;
}

void not_implemented(int fd) {
	http_reply(fd, NULL, 501, "Not Implemented", "text/plain", "That command is not implemented");
}

void do_404(char *item, int fd) {
	http_reply(fd, NULL, 404, "Not Found", "text/plain", "The item you seek is not here");
}

int isadir(char *f) {
	struct stat	info;
	
	return (stat(f, &info) != -1 && S_ISDIR(info.st_mode));
}

char *split(char *arg, char c) {
	char	*cmd;
	
	if ((cmd = strchr(arg, c)) != NULL) {
		arg[strlen(arg) - strlen(cmd)] = '\0';
		*cmd++;
		return cmd;
	}
	return "";
}

int not_exist(char *f) {
	struct stat info;
		
	strcpy(cmdstring,split(f, '?'));
	return (stat(f, &info) == -1);
}

void getowner(char *f) {
	struct stat info;
	char *name = (char *) malloc(sizeof(f));
	char *path = (char *) malloc(sizeof(f));
	
	if ((name = strrchr(f, '/')) != NULL) {
		name++;
		strcpy(path, f);
		path[strlen(f) - strlen(name)] = '\0';
	}
	if (strncmp(f, "./", 2) == 0)
		*path++;
	if (stat(f, &info) != -1) {
		optr.uid = info.st_uid;
		optr.gid = info.st_gid;
		optr.Path = malloc(sizeof(char) * strlen(path));
		optr.Name = malloc(sizeof(char) * strlen(name));
		optr.Path = path;
		optr.Name = name;
	}
}
	
int isexec(char *f) {
	struct stat info;

	if (stat(f, &info) != -1 && S_ISREG(info.st_mode)) {
		if (info.st_uid == optr.uid || info.st_gid == optr.gid) {
			if (info.st_uid == optr.uid)
				return ((info.st_mode & ~(S_IFMT)) & S_IXUSR);
			if (info.st_gid == optr.gid)
				return ((info.st_mode & ~(S_IFMT)) & S_IXGRP);
		}
		else
			return ((info.st_mode & ~(S_IFMT)) & S_IXOTH);
	}
	return 0;
}

void skip_rest_of_header(FILE *fp) {
	char	buf[BUFSIZ];
	
	while(fgets(buf, BUFSIZ, fp) != NULL && strcmp(buf, "\r\n") != 0)
		;
}

int check_rfc1945(char *cmd) {
	int		i;
	
	for (i = 0; i < sizeof(rfc1945_commands); i++)
		if (strcmp(cmd, rfc1945_commands[i]) == 0)
			return i;
	return -1;
}

void do_head(int fd) {
	server_bytes_sent += http_reply(fd, NULL, 200, "OK", "text/plain", NULL);
}

int is_cgi(char *f) {
	return isexec(f);
}

char *file_type(char *f) {
	char	*cp;
	
	if ((cp = strrchr(f, '.')) != NULL)
		return cp + 1;
	return "";
}

void deamonize(void) {
	/* Fork off the parent process */
	pid = fork();
	if (pid < 0) {
		errorlog("Terminating: Error, fork not possible");
		exit(EXIT_FAILURE);
	}
	/* If we got a good PID, then
	   we can exit the parent process. */
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	/* Change the file mode mask */
	umask(0);
		
	/* Open any logs here */	
		
	/* Create a new SID for the child process */
	sid = setsid();
	if (sid < 0) {
		/* Log the failure */
		errorlog("Terminating: Error, fialed to get new sid to child");
		exit(EXIT_FAILURE);
	}
	
	infolog("Info: Switching to deamon mode");
}

void infolog(char *msg) {
	if (deamon) {
		openlog("webserver", 0, LOG_USER);
		syslog(LOG_INFO, "%s", msg);
		closelog();
	}
	if (debug)
		fprintf(stderr, "%s\n", msg);
}

void errorlog(char *msg) {
	if (deamon) {
		openlog("webserver", 0, LOG_USER);
		syslog(LOG_ALERT, "%s", msg);
		closelog();
	}
	if (debug)
		fprintf(stderr, "%s\n", msg);
	exit(EXIT_FAILURE);
}
