/***************************************************************************
 *            workers.c
 *
 *  Tue Jun 29 19:47:41 2004
 *  Copyright  2004  Michael Rasmussen
 *  mir@miras.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "utils.h"
#include "workers.h"

void do_ls(char *dir, int fd) {
	DIR				*dirptr;
	struct dirent	*direntp;
	FILE			*fp;
	int				bytes = 0;
	char			path[stringbuf] = "/";
	
	fp = fdopen(fd, "w");
	bytes = http_reply(fd, &fp, 200, "OK", "text/html", NULL);
	bytes += fprintf(fp, "<h2 align=\"center\">Listing of Directory %s</h2>\n", dir);

	if ((dirptr = opendir(dir)) != NULL) {
		if (strcmp(dir, ".") != 0)
			strcat(path, dir);	
		while ((direntp = readdir(dirptr))) {
			if (strcmp(direntp->d_name, ".") == 0)
				continue;
			bytes += fprintf(fp, "<a href=\"%s/%s\">%s</a><br />\n",
						(strcmp(".", dir) == 0)? "": path,
						direntp->d_name,
						direntp->d_name
					);
		}
		closedir(dirptr);
	}
	fclose(fp);
	server_bytes_sent += bytes;
}

void do_cat(char *f, int fd) {
	char	*extension = file_type(f);
	char	*type = (char *) malloc(sizeof(int) * 25);
	int		bytes = 0;
	int		in, out, maxfd;
	int		io[2];
	fd_set	watchsetread, watchsetwrite;
	fd_set	insetread, insetwrite;
	char 	buf[4096];
	int		writeflag = 0;
	
	in = out = maxfd = 0;
	if (strcmp(extension, "html") == 0)
		strcpy(type, "text/html");
	else if (strcmp(extension, "gif") == 0)
		strcpy(type, "image/gif");
	else if (strcmp(extension, "jpg") == 0)
		strcpy(type, "image/jpeg");
	else if (strcmp(extension, "jpeg") == 0)
		strcpy(type, "image/jpeg");
	else if (strcmp(extension, "png") == 0)
		strcpy(type, "image/png");
	else
		strcpy(type, "text/plain");
	
	if ((io[0] = open(f, O_RDONLY)) < 0) { // read
		errorlog("Terminating: Error open input file");
	}
	io[1] = fd; // write
	
	strcpy(buf, standardHeader(200, "OK", type));
	server_bytes_sent += write(io[1], buf, strlen(buf));
	FD_ZERO(&watchsetread);
	FD_ZERO(&watchsetwrite);
	FD_SET(io[0], &watchsetread);
	FD_SET(io[1], &watchsetwrite);
	maxfd = io[0] > io[1]? io[0]: io[1];
	while (FD_ISSET(io[0], &watchsetread) &&
			FD_ISSET(io[1], &watchsetwrite)) {
		insetread = watchsetread;
		insetwrite = watchsetwrite;
		if (select(maxfd+1, &insetread, &insetwrite, NULL, NULL) < 0) {
			errorlog("Terminating: Error watching IO files");
		}
		if (FD_ISSET(io[0], &insetread) && !writeflag) {
			in = read(io[0], buf, sizeof(buf));
			if (in < 0) {
				errorlog("Terminating: Error reading input");
			}
			else if (!in) {
				close(io[0]);
				FD_CLR(io[0], &watchsetread);
			}
			else
				writeflag = 1;
		}
		if (FD_ISSET(io[1], &insetwrite) && writeflag) {
			out = write(io[1], buf, in);
			if (out < 0) {
				errorlog("Terminating: Error writing to output");
			}
			else if (!out) {
				close(io[1]);
				FD_CLR(io[1], &watchsetwrite);
			}
			else {
				writeflag = 0;
				bytes += out;
			}
		}
	}
	free(type);
	server_bytes_sent += bytes;
}

int built_in(char *arg, int fd) {
	FILE	*fp;
	
	if (strcmp(arg, "status") != 0)
		return 0;
	fp = fdopen(fd, "w");
	http_reply(fd, &fp, 200, "OK", "text/plain", NULL);
	fprintf(fp, "Server started: %s", ctime(&server_started));
	fprintf(fp, "Total requests: %d\n", server_requests);
	fprintf(fp, "Bytes sent out: %d\n", server_bytes_sent);
	fclose(fp);
	return 1;
}

void do_exec(char *prog, int fd) {
	pid_t	pid;
	FILE	*fpsock;
	int		i, fdin;
	char 	program[BUFSIZ];
	char 	*argv[100];
	char 	curpath[1024];
	char 	*tmp;
	
	
	getcwd(curpath, sizeof(curpath));
	strcpy(program, curpath);
	strcat(program, optr.Path);
	strcat(program, prog);
	argv[0] = program;
	i = 1;
	pthread_mutex_lock (&parse_command_line);
	while ((tmp = split(cmdstring, '&')) != "") {
		argv[i] = malloc(sizeof(char) * strlen(cmdstring)+1);
		memcpy(argv[i++], cmdstring, strlen(cmdstring)+1);
		strcpy(cmdstring,tmp);
	}
	if (cmdstring && i > 1) {
		argv[i] = malloc(sizeof(char) * strlen(cmdstring)+1);
		memcpy(argv[i++], cmdstring, strlen(cmdstring)+1);
	}
	pthread_mutex_unlock(&parse_command_line);
	argv[i] = NULL;
	i = 0;
	switch (pid = fork()) {
		case -1:
			errorlog("Terminating: Fork returned error code, no child");
			break;
		case 0: 
			fpsock = fdopen(fd, "w");
			if ((fdin = open(program, O_RDONLY)) < 0)
				errorlog("Terminating: Error open file for input");
			
			if (dup2(fd, STDOUT_FILENO) == -1)
				exit(1);
			if (dup2(fdin, STDIN_FILENO) == -1)
				exit(1);
			http_reply(fd, &fpsock, 200, "OK", NULL, NULL);
			execvp(program, argv);
			break;
		default:
			wait(NULL);
			kill(pid, SIGTERM);
	}
	return;
}

void process_rq(char *rq, int fd) {
	char	cmd[BUFSIZ], arg[BUFSIZ];
	
	if (sscanf(rq, "%s%s", cmd, arg) != 2)
		return;
	sanitize(arg);
	
	if ((rfc1945_command = check_rfc1945(cmd)) == -1) {
		not_implemented(fd);
		return;
	}
	
	switch (rfc1945_command) {
		case 0:
			do_head(fd);
			break;
		case 1:
			if (built_in(arg, fd))
				;
			else if (not_exist(arg))
				do_404(arg, fd);
			else if (isadir(arg))
				do_ls(arg, fd);
			else if (is_cgi(arg))
				do_exec(arg, fd);
			else
				do_cat(arg, fd);
			break;
		default:
			errorlog("Terminating: Error, unrecognized command");
	}
	return;
}

void *handle_call(void *fdptr) {
	FILE	*fpin;
	char	request[BUFSIZ];
	int		fd;
	
	fd = *(int *) fdptr;
	
	fpin = fdopen(fd, "r");
	fgets(request, BUFSIZ, fpin);
	skip_rest_of_header(fpin);
	
	process_rq(request, fd);
	
	fclose(fpin);
	return (void *) NULL;
}
